#include "Pathfinder.h"
#include "Heuristic.h"
#include <memory>
#include <sstream>
#include <utility>

class AStar: public Pathfinder {
public:
	void preprocess(std::vector<int> &map, int width, int height){
		chooseRoots(map, width, height, nRoots, RootStrategy::FAR);
		heuristic->preprocess(map, width, height, roots);
	}

	void findPath(Position start, Position goal, std::vector<Position> &path,
			std::vector<int> &map, int width, int height);

	void clear(){
		heuristic->clear();
	}

	std::string getName(){
		return name;
	}

	size_t getSize(){
		return sizeof(double) * distances.size() + sizeof(int) * parents.size() + heuristic->getSize();
	}

	AStar(int nRoots, std::unique_ptr<Heuristic> h) :
			nRoots(nRoots), heuristic(std::move(h)){
		std::stringstream stream;
		stream << "AStar-" << heuristic->getName() << nRoots;
		name = stream.str();
	}

private:
	int nRoots;
	std::unique_ptr<Heuristic> heuristic;
	std::vector<double> distances;
	std::vector<int> parents;
	std::string name;
};
