#ifndef LCATREE_H
#define LCATREE_H
#include <vector>
#include <cmath>
#include <cstddef>

class LcaTree {

public:
	LcaTree(int root, int size) :
			root(root), parents(size, -1), tour(2 * size - 1, -1), levels(2 * size - 1), representatives(
					size, -1), mins(ceil(log2(2 * size - 1))){
	}

	void addNode(int parentIndex, int nodeIndex){
		parents[nodeIndex] = parentIndex;
	}

	int getParent(int index){
		return parents[index];
	}

	size_t getSize(){
		size_t size = parents.size() + tour.size() + levels.size() + representatives.size();
		for(auto vec : mins){
			size += vec.size();
		}
		return size * sizeof(int);
	}

	void prepareLCA();
	int getLCA(int indexA, int indexB);

private:
	//the tree
	int root;
	std::vector<int> parents;
	//LCA -> RMQ
	std::vector<int> tour;
	std::vector<int> levels;
	std::vector<int> representatives;
	//RMQ
	std::vector<std::vector<int> > mins;

	std::vector<std::vector<int> > getChildren();
	void euler(int nodeIndex, int level, int &arrayIndex, std::vector<std::vector<int> > &children);
	void prepareMins();
	int rmq(int i, int j);
};
#endif //LCATREE_H
