#include "AStar.h"
#include <queue>

#ifdef VISUAL
#include <functional>
#endif

void AStar::findPath(Position start, Position goal, std::vector<Position> &path,
		std::vector<int> &map, int width, int height){
	distances.clear();
	distances.resize(map.size(), std::numeric_limits<double>::infinity());
	parents.clear();
	parents.resize(map.size(), -1);
	std::priority_queue<std::pair<double, int>, std::vector<std::pair<double, int> >,
			std::greater<std::pair<double, int> > > queue;

	int startIndex = start.toIndex(width);
	int goalIndex = goal.toIndex(width);
	distances[startIndex] = 0;
	queue.push(std::make_pair(heuristic->getValue(startIndex, goalIndex), startIndex));

	while(!queue.empty()){
		std::pair<double, int> node = queue.top();
		queue.pop();
		if(node.second == goalIndex){
			int nextNode = goalIndex;
			while(nextNode != -1){
				path.push_back(Position::fromIndex(nextNode, width));
				nextNode = parents[nextNode];
			}
			std::reverse(path.begin(), path.end());
			return;
		}
		if(node.first > distances[node.second] + heuristic->getValue(node.second, goalIndex) + EPS)
			continue;

		auto neighbours = getNeighbours(Position::fromIndex(node.second, width), map, width,
				height);
		for(std::pair<Position, double> neighbour : neighbours){
			int nIndex = neighbour.first.toIndex(width);
			double dist = distances[node.second] + neighbour.second;
			if(dist < distances[nIndex]){
				distances[nIndex] = dist;
				parents[nIndex] = node.second;
				queue.push(std::make_pair(dist + heuristic->getValue(nIndex, goalIndex), nIndex));
			}
		}
	}
}
