#include "LcaTree.h"

#ifdef VISUAL
#include <intrin.h>
#pragma intrinsic(_BitScanReverse)
#endif

void LcaTree::prepareLCA(){
	std::vector<std::vector<int> > children = getChildren();
	int a = 0;
	euler(root, 0, a, children);
	prepareMins();
}

int LcaTree::getLCA(int indexA, int indexB){
	int repA = representatives[indexA];
	int repB = representatives[indexB];
	if(repA == -1 || repB == -1) //nodes not connected to the root
		return -1;
	return repA < repB ? tour[rmq(repA, repB)] : tour[rmq(repB, repA)];
}

int LcaTree::rmq(int i, int j){
#ifndef VISUAL
	int k = 31 - __builtin_clz(j - i + 1); //log2, truncated
#else
	unsigned long index;
	_BitScanReverse(&index, j - i + 1); //log2, truncated
	int k = index;
#endif
	int pow2k = 1 << k; //2^k
	return levels[mins[k][i]] < levels[mins[k][j - pow2k + 1]] ? mins[k][i] : mins[k][j - pow2k + 1];
}

std::vector<std::vector<int> > LcaTree::getChildren(){
	std::vector<std::vector<int> > children(parents.size());
	for(unsigned int i = 0; i < parents.size(); i++){
		if(parents[i] == -1)
			continue;
		children[parents[i]].push_back(i);
	}
	return children;
}

void LcaTree::euler(int nodeIndex, int level, int &arrayIndex,
		std::vector<std::vector<int> > &children){
	tour[arrayIndex] = nodeIndex;
	levels[arrayIndex] = level;
	for(int childIndex : children[nodeIndex]){
		arrayIndex++;
		euler(childIndex, level + 1, arrayIndex, children);
		arrayIndex++;
		tour[arrayIndex] = nodeIndex;
		levels[arrayIndex] = level;
	}
}

void LcaTree::prepareMins(){
	for(int i = tour.size() - 1; i >= 0; i--){
		if(tour[i] == -1)
			continue;
		representatives[tour[i]] = i;
	}
	mins[0] = std::vector<int>(levels.size());
	for(unsigned int j = 0; j < mins[0].size(); j++){
		mins[0][j] = j;
	}
	int pow2i = 1; //2^i, i=0
	int pow2im1 = -1; //2^(i-1)
	for(unsigned int i = 1; i < mins.size(); i++){
		pow2im1 = pow2i;
		pow2i *= 2;
		mins[i] = std::vector<int>(levels.size() - pow2i + 1);
		for(unsigned int j = 0; j < mins[i].size(); j++){
			mins[i][j] =
					levels[mins[i - 1][j]] <= levels[mins[i - 1][j + pow2im1]] ?
							mins[i - 1][j] : mins[i - 1][j + pow2im1];
		}
	}
}
