#include "MultipleTrees.h"
#include <queue>
#include <limits>
#include <deque>

#ifdef VISUAL
#include <functional>
#endif

void MultipleTrees::preprocess(std::vector<int> &map, int width, int height){
	parents.clear();

	std::priority_queue<std::pair<double, int>, std::vector<std::pair<double, int> >,
			std::greater<std::pair<double, int> > > queue;

	chooseRoots(map, width, height, nRoots, RootStrategy::GRID);

//	nRoots = roots.size();
//	std::stringstream stream;
//	stream << "MultipleTrees" << nRoots;
//	name = stream.str();

	for(Position root : roots){
		std::vector<std::pair<double, int> > distParents;
		distParents.resize(map.size(), std::make_pair(std::numeric_limits<double>::infinity(), -1));

		distParents[root.toIndex(width)].first = 0;
		queue.push(std::make_pair(0, root.toIndex(width)));

		while(!queue.empty()){
			std::pair<double, int> node = queue.top();
			queue.pop();
			if(node.first > distParents[node.second].first + EPS)
				continue;

			auto neighbours = getNeighbours(Position::fromIndex(node.second, width), map, width,
					height);
			for(std::pair<Position, double> neighbour : neighbours){
				double dist = node.first + neighbour.second;
				int nIndex = neighbour.first.toIndex(width);
				if(dist < distParents[nIndex].first){
					distParents[nIndex].first = dist;
					distParents[nIndex].second = node.second;
					queue.push(std::make_pair(dist, nIndex));
				}
			}
		}
		parents.push_back(distParents);
	}
}

void MultipleTrees::findPath(Position start, Position goal, std::vector<Position> &path,
		std::vector<int> &map, int width, int height){
	double tempDist, minDist = std::numeric_limits<double>::infinity();
	int minTree = -1;
	int startIndex = start.toIndex(width);
	int goalIndex = goal.toIndex(width);
	for(unsigned int i = 0; i < parents.size(); i++){
		tempDist = parents[i][startIndex].first + parents[i][goalIndex].first;
		if(tempDist < minDist){
			minTree = i;
			minDist = tempDist;
		}
	}
	if(minTree == -1){ //path impossible
		return;
	}
	std::deque<Position> rootToGoal;
	int nextNode = startIndex;
	while(parents[minTree][nextNode].second != -1){
		path.push_back(Position::fromIndex(nextNode, width));
		nextNode = parents[minTree][nextNode].second;
	}
	nextNode = goalIndex;
	while(nextNode != -1){
		rootToGoal.push_front(Position::fromIndex(nextNode, width));
		nextNode = parents[minTree][nextNode].second;
	}
	for(Position pos : rootToGoal){
		path.push_back(pos);
	}
}
