#include "SimpleLCA.h"
#include <queue>
#include <limits>
#include <deque>

#ifdef VISUAL
#include <functional>
#endif

void SimpleLCA::preprocess(std::vector<int> &map, int width, int height){
	std::vector<double> distances(map.size(), std::numeric_limits<double>::infinity());

	std::priority_queue<std::pair<double, int>, std::vector<std::pair<double, int> >,
			std::greater<std::pair<double, int> > > queue;

	chooseRoots(map, width, height, 1, RootStrategy::GRID);
	int rootIndex = roots[0].toIndex(width);
	parents.clear();
	parents.resize(map.size(), std::make_pair(-1, false));

	distances[rootIndex] = 0;
	queue.push(std::make_pair(0, rootIndex));

	while(!queue.empty()){
		std::pair<double, int> node = queue.top();
		queue.pop();
		if(node.first > distances[node.second] + EPS)
			continue;

		auto neighbours = getNeighbours(Position::fromIndex(node.second, width), map, width,
				height);
		for(std::pair<Position, double> neighbour : neighbours){
			double dist = node.first + neighbour.second;
			int nIndex = neighbour.first.toIndex(width);
			if(dist < distances[nIndex]){
				distances[nIndex] = dist;
				parents[nIndex].first = node.second;
				queue.push(std::make_pair(dist, nIndex));
			}
		}
	}
}

void SimpleLCA::findPath(Position start, Position goal, std::vector<Position> &path,
		std::vector<int> &map, int width, int height){
	std::deque<Position> rootToGoal;
	int nextNode = start.toIndex(width);
	while(parents[nextNode].first != -1){
		path.push_back(Position::fromIndex(nextNode, width));
		parents[nextNode].second = true;
		nextNode = parents[nextNode].first;
	}
	int startRoot = nextNode;
	nextNode = goal.toIndex(width);
	while(parents[nextNode].first != -1){
		if(parents[nextNode].second){ //LCA
			for(Position pos : path){
				parents[pos.toIndex(width)].second = false;
			}
			auto pathIterator = path.begin();
			while(pathIterator->toIndex(width) != nextNode)
			pathIterator++;
			path.erase(pathIterator, path.end());
			startRoot = nextNode;
			break;
		}
		rootToGoal.push_front(Position::fromIndex(nextNode, width));
		nextNode = parents[nextNode].first;
	}
	for(Position pos : path){
		parents[pos.toIndex(width)].second = false;
	}
	if(startRoot != nextNode){ //path impossible
		path.clear();
		return;
	}
	path.push_back(Position::fromIndex(startRoot, width));
	for(Position pos : rootToGoal){
		path.push_back(pos);
	}
}
