#include "TreeCacheImproved.h"
#include <queue>
#include <limits>
#include <deque>

#ifdef VISUAL
#include <functional>
#endif

void TreeCacheImproved::preprocess(std::vector<int> &map, int width, int height){
	trees.clear();
	distancesPerTree.clear();

	std::priority_queue<std::pair<double, int>, std::vector<std::pair<double, int> >,
			std::greater<std::pair<double, int> > > queue;

	chooseRoots(map, width, height, nRoots, RootStrategy::GRID);

//	nRoots = roots.size();
//	std::stringstream stream;
//	stream << "TreeCacheImproved" << nRoots;
//	name = stream.str();

	for(Position root : roots){
		LcaTree tree(root.toIndex(width), map.size());
		std::vector<double> distances(map.size(), std::numeric_limits<double>::infinity());

		distances[root.toIndex(width)] = 0;
		queue.push(std::make_pair(0, root.toIndex(width)));

		while(!queue.empty()){
			std::pair<double, int> node = queue.top();
			queue.pop();
			if(node.first > distances[node.second] + EPS)
				continue;

			auto neighbours = getNeighbours(Position::fromIndex(node.second, width), map, width,
					height);
			for(std::pair<Position, double> neighbour : neighbours){
				double dist = node.first + neighbour.second;
				int nIndex = neighbour.first.toIndex(width);
				if(dist < distances[nIndex]){
					distances[nIndex] = dist;
					tree.addNode(node.second, nIndex);
					queue.push(std::make_pair(dist, nIndex));
				}
			}
		}
		tree.prepareLCA();
		trees.push_back(tree);
		distancesPerTree.push_back(distances);
	}
}

void TreeCacheImproved::findPath(Position start, Position goal, std::vector<Position> &path,
		std::vector<int> &map, int width, int height){
	int minTreeIndex = -1;
	int minLcaIndex = -1;
	double minDistance = std::numeric_limits<double>::infinity();
	int startIndex = start.toIndex(width);
	int goalIndex = goal.toIndex(width);
	for(unsigned int i = 0; i < trees.size(); i++){
		int lcaIndex = trees[i].getLCA(startIndex, goalIndex);
		if(lcaIndex == -1)
			continue;
		std::vector<double>& distances = distancesPerTree[i];
		double distance = distances[startIndex] + distances[goalIndex] - 2 * distances[lcaIndex];
		if(distance < minDistance){
			minDistance = distance;
			minLcaIndex = lcaIndex;
			minTreeIndex = i;
		}
	}
	if(minTreeIndex == -1) //path impossible
		return;

	LcaTree& minTree = trees[minTreeIndex];
	std::deque<Position> LcaToGoal;
	int nextNode = startIndex;
	while(nextNode != minLcaIndex){
		path.push_back(Position::fromIndex(nextNode, width));
		nextNode = minTree.getParent(nextNode);
	}
	nextNode = goalIndex;
	while(nextNode != minLcaIndex){
		LcaToGoal.push_front(Position::fromIndex(nextNode, width));
		nextNode = minTree.getParent(nextNode);
	}
	path.push_back(Position::fromIndex(minLcaIndex, width));
	for(Position pos : LcaToGoal){
		path.push_back(pos);
	}
}
